unit ListUnit;

interface

uses
  Classes,
  SysUtils,
  RetainableUnit;

type

  TListSortCompareMethod = function(Item1, Item2: Pointer): integer of object;

  IIRetainableGenericList = interface(IRetainable)
    ['{2FB42930-7D83-401F-9FF9-DC3AA5B3340B}']
    procedure Clear;
    procedure DeleteIndex(AIndex: integer);
    procedure ClearAndFree;
  end;

  TGenericListBase = class(TRetainable)
  strict protected
    function GetCount: integer; virtual; abstract;
    procedure SetCount(ANewCount: integer); virtual; abstract;

  protected

    //{$IFDEF MSWINDOWS}
    //  Protected
    //{$ELSE}
  public
    //{$ENDIF}
    FData: TList;

  public
    function InternalGet(AIndex: integer): Pointer; virtual; abstract;

  published
    property Count: integer read GetCount write SetCount;
  end;

  // Although this is interface list, it does not autoincerement refcounts - use TIRetainableGenericList instead
  TIGenericList<TItem{$IFDEF DCC}: IInterface{$ENDIF}> = class(TGenericListBase)
  strict private
    OwnedList: boolean;

    function GetCapacity: integer;
    procedure SetCapacity(ANewCapacity: integer);

    {$IFDEF FPC}
    procedure QuickSort(FList: TList; L, R: longint; Compare: TListSortCompareMethod);
    {$ENDIF}
  strict protected
    function GetCount: integer; override;
    procedure SetCount(ANewCount: integer); override;
    function Get(AIndex: integer): TItem; virtual;
    procedure Put(AIndex: integer; AItem: TItem); virtual;

  public
    {$IFNDEF PASCAL_SCRIPT}
    function InternalGet(AIndex: integer): Pointer; override;
    constructor CreateWrapperOf(AWrappedList: TGenericListBase); overload;
    constructor CreateWrapperOf(AWrappedList: TList); overload;
    {$ENDIF}

    constructor Create;
    destructor Destroy; override;

    function Add(AItem: TItem): integer; virtual;
    procedure Clear; virtual;

    procedure DeleteIndex(AIndex: integer); virtual;
    procedure Delete(AItem: TItem); virtual;

    procedure Exchange(AIndex1, AIndex2: integer); virtual;
    procedure Expand;
    function Extract(AItem: TItem): TItem; virtual;
    function First: TItem;
    function IndexOf(AItem: TItem): integer;
    procedure Insert(AIndex: integer; AItem: TItem); virtual;
    function Last: TItem;
    procedure Move(ACurIndex, ANewIndex: integer); virtual;
    function Remove(AItem: TItem): integer; virtual;
    procedure Pack;
    procedure Sort(ACompare: TListSortCompare); virtual;
    procedure SortByMethod(ACompareMethod: TListSortCompareMethod); virtual;

    property Capacity: integer read GetCapacity write SetCapacity;
    property Items[Index: integer]: TItem read Get write Put; default;
  end;

  // Refcounts maintained by the list
  TIRetainableGenericList<TItem{$IFDEF DCC}: IRetainable{$ENDIF}> = class(TIGenericList<TItem>, IIRetainableGenericList)
  strict private
  strict protected
    function Get(AIndex: integer): TItem; override;
    procedure Put(AIndex: integer; AItem: TItem); override;
    procedure SetCount(ANewCount: integer); override;

  public
    destructor Destroy; override;

    function Add(AItem: TItem): integer; override;
    function Push(AItem: TItem): integer;
    procedure Clear; override;
    procedure DeleteIndex(AIndex: integer); override;
    procedure Delete(AItem: TItem); override;
    procedure Insert(AIndex: integer; AItem: TItem); override;
    function Remove(AItem: TItem): integer; override;
    procedure ClearAndFree;

    property Items[Index: integer]: TItem read Get write Put; default;
    function GetRetainableObject(AIndex: integer): TRetainable;
  end;

implementation

destructor TIRetainableGenericList<TItem>.Destroy;
begin
  ClearAndFree;

  inherited;
end;

procedure TIRetainableGenericList<TItem>.ClearAndFree;
var
  I: longint;
  T: TRetainable;
begin
  for I := 0 to Count - 1 do
  begin
    T := GetRetainableObject(I);
    TRetainable.ReleaseAndNil(T);
  end;

  inherited Clear;
end;

function TIRetainableGenericList<TItem>.Get(AIndex: integer): TItem;
begin
  Result := inherited Get(AIndex);
end;

function TIRetainableGenericList<TItem>.GetRetainableObject(AIndex: integer): TRetainable;
var
  X: IRetainable;
begin
  X := Self[AIndex];
  Result := X as TRetainable;
end;

procedure TIRetainableGenericList<TItem>.Put(AIndex: integer; AItem: TItem);
var
  X: IRetainable;
begin
  if (AIndex < Count) then
  begin
    X := Self[AIndex];
    TRetainable.Release(X as TRetainable);
  end;
  IRetainable(AItem).Retain;
  inherited Put(AIndex, AItem);
end;

procedure TIRetainableGenericList<TItem>.SetCount(ANewCount: integer);
var
  I: integer;
  X: IRetainable;
begin
  // release objects when the list is beign shortened
  for I := ANewCount to Count - 1 do
  begin
    X := Self[I];
    TRetainable.Release(X as TRetainable);
  end;

  inherited SetCount(ANewCount);
end;

function TIRetainableGenericList<TItem>.Add(AItem: TItem): integer;
var
  X: IRetainable;
begin
  X := AItem;
  (X as TRetainable).Retain;

  Result := inherited Add(AItem);
end;

function TIRetainableGenericList<TItem>.Push(AItem: TItem): integer;
begin
  Result := inherited Add(AItem);
end;

procedure TIRetainableGenericList<TItem>.Clear;
begin
  ClearAndFree;
end;

procedure TIRetainableGenericList<TItem>.DeleteIndex(AIndex: integer);
var
  X: IRetainable;
begin
  if AIndex < Count then
  begin
    X := Self[AIndex];
    TRetainable.Release(X as TRetainable);
  end;

  inherited DeleteIndex(AIndex);
end;

procedure TIRetainableGenericList<TItem>.Delete(AItem: TItem);
begin
  Remove(AItem);
end;

procedure TIRetainableGenericList<TItem>.Insert(AIndex: integer; AItem: TItem);
var
  X: IRetainable;
begin
  X := AItem;
  (X as TRetainable).Retain;
  inherited Insert(AIndex, AItem);
end;

function TIRetainableGenericList<TItem>.Remove(AItem: TItem): integer;
var
  I: integer;
begin
  I := IndexOf(AItem);
  if I >= 0 then
    DeleteIndex(I);

  Result := I;
end;


constructor TIGenericList<TItem>.Create;
begin
  inherited;

  FData := TList.Create;
  OwnedList := True;
end;

constructor TIGenericList<TItem>.CreateWrapperOf(AWrappedList: TGenericListBase);
begin
  inherited;

  FData := AWrappedList.FData;
  OwnedList := False;
end;

constructor TIGenericList<TItem>.CreateWrapperOf(AWrappedList: TList);
begin
  inherited;

  FData := AWrappedList;
  OwnedList := False;
end;

destructor TIGenericList<TItem>.Destroy;
begin
  if OwnedList then
    FreeNil(FData);
  inherited;
end;

function TIGenericList<TItem>.Add(AItem: TItem): integer;
begin
  Result := FData.Add(PPointer(@AItem)^);
end;

procedure TIGenericList<TItem>.Clear;
begin
  FData.Clear;
end;

procedure TIGenericList<TItem>.DeleteIndex(AIndex: integer);
begin
  FData.Delete(AIndex);
end;

procedure TIGenericList<TItem>.Exchange(AIndex1, AIndex2: integer);
begin
  FData.Exchange(AIndex1, AIndex2);
end;

procedure TIGenericList<TItem>.Expand;
begin
  FData.Expand;
end;

function TIGenericList<TItem>.Extract(AItem: TItem): TItem;
type
  PItem = ^TItem;
var
  P: Pointer;
begin
  P := FData.Extract(PPointer(@AItem)^);
  Result := (PItem(@P))^;
end;

function TIGenericList<TItem>.First: TItem;
type
  PItem = ^TItem;
var
  P: Pointer;
begin
  {$IFDEF DCC}
  if FData.Count = 0 then
    exit(nil);
  {$ENDIF}

  P := FData.First;
  Result := (PItem(@P))^;
end;

function TIGenericList<TItem>.IndexOf(AItem: TItem): integer;
begin
  Result := FData.IndexOf(PPointer(@AItem)^);
end;

procedure TIGenericList<TItem>.Insert(AIndex: integer; AItem: TItem);
begin
  FData.Insert(AIndex, PPointer(@AItem)^);
end;

function TIGenericList<TItem>.Last: TItem;
type
  PItem = ^TItem;
var
  P: Pointer;
begin
  if Count = 0 then
    Result := nil
  else
  begin
    P := FData.Last;
    Result := (PItem(@P))^;
  end;
end;

procedure TIGenericList<TItem>.Move(ACurIndex, ANewIndex: integer);
begin
  FData.Move(ACurIndex, ANewIndex);
end;

procedure TIGenericList<TItem>.Delete(AItem: TItem);
begin
  FData.Remove(PPointer(@AItem)^);
end;

function TIGenericList<TItem>.Remove(AItem: TItem): integer;
begin
  Result := FData.Remove(PPointer(@AItem)^);
end;

procedure TIGenericList<TItem>.Pack;
begin
  FData.Pack;
end;

procedure TIGenericList<TItem>.Sort(ACompare: TListSortCompare);
begin
  FData.Sort(@ACompare);
end;

{$IFDEF FPC}
procedure TIGenericList<TItem>.QuickSort(FList: TList; L, R: longint; Compare: TListSortCompareMethod);
var
  I, J: longint;
  P, Q: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := FList[(L + R) div 2];
    repeat
      while Compare(P, FList[i]) > 0 do
      begin
        I := I + 1;
      end;
      while Compare(P, FList[J]) < 0 do
      begin
        J := J - 1;
      end;
      if I <= J then
      begin
        Q := FList[I];
        Flist[I] := FList[J];
        FList[J] := Q;
        I := I + 1;
        J := J - 1;
      end;
    until I > J;
    // sort the smaller range recursively
    // sort the bigger range via the loop
    // Reasons: memory usage is O(log(n)) instead of O(n) and loop is faster than recursion
    if J - L < R - I then
    begin
      if L < J then
        QuickSort(FList, L, J, Compare);
      L := I;
    end
    else
    begin
      if I < R then
        QuickSort(FList, I, R, Compare);
      R := J;
    end;
  until L >= R;
end;

{$ENDIF}


procedure TIGenericList<TItem>.SortByMethod(ACompareMethod: TListSortCompareMethod);
begin
  {$IFDEF DCC}
  FData.SortList(
    function (Item1, Item2: Pointer): integer
    begin
      Result := ACompareMethod(Item1, Item2);
    end);
  {$ELSE}
  if FData.Count < 2 then
    exit;
  QuickSort(FData, 0, FData.Count - 1, ACompareMethod);
  {$ENDIF}
end;

function TIGenericList<TItem>.Get(AIndex: integer): TItem;
type
  PItem = ^TItem;
var
  P: Pointer;
begin
  P := FData[AIndex];
  Result := (PItem(@P))^;
end;

procedure TIGenericList<TItem>.Put(AIndex: integer; AItem: TItem);
begin
  FData[AIndex] := PPointer(@AItem)^;
end;

function TIGenericList<TItem>.GetCapacity: integer;
begin
  Result := FData.Capacity;
end;

function TIGenericList<TItem>.GetCount: integer;
begin
  Result := FData.Count;
end;

procedure TIGenericList<TItem>.SetCapacity(ANewCapacity: integer);
begin
  FData.Capacity := ANewCapacity;
end;

procedure TIGenericList<TItem>.SetCount(ANewCount: integer);
begin
  FData.Count := ANewCount;
end;

function TIGenericList<TItem>.InternalGet(AIndex: integer): Pointer;
begin
  {$IFDEF FPC}
  Result := nil;
  AIndex := AIndex;
  {$ENDIF}
  raise Exception.Create('unsupported operation');
end;

end.
